/* vm7.h -     definitions/declarations for the memory vm7 handler
 */

#ifndef VM7_H
#define VM7_H

#ifdef DEBUG
#ifndef VCMPRS
#define VCMPRS
#endif
#endif

#include "stypes.h"
#include "stdio.h"
#include "smem.h"

typedef ULONG VM7HDL;

#define VM7NULLHDL	0L
#define MAXBANK      7
#define DEFAULT_BANK 1

#define READ_WRITE   1
#define READ_ONLY    2
#define BMASK        0x1fffffff
                        /*  used to mask out bank id from handle  */

#define MIN_BLOCK_SIZE   16             /* minimum block size allocated */
#define END_FREE_LIST    0x0fffffff
#define INVALID_PAGE     0x7fff
#define PAGED_OUT        0x4000
#define PAGED_EXT        0x8000
#define PAGE_NUMBER      0x3fff

#ifndef NULL_ITEM
#define NULL_ITEM 0xffffffff
#endif

#ifndef __SWAP__
struct Vaddress
   {
   unsigned offset: 15;
   unsigned lock  :  1;
   unsigned index : 12;
   unsigned vl    :  1;
   unsigned id    :  3;
   };
#else
struct Vaddress
   {
   unsigned id    :  3;
   unsigned vl    :  1;
   unsigned index : 12;
   unsigned lock  :  1;
   unsigned offset: 15;
   };
#endif
typedef struct Vaddress VADDRESS;

union VHandle
   {
   VADDRESS a;
   VM7HDL   h;
   };
typedef union VHandle VVM7HDL;

struct HandleList
   {
   VADDRESS     address;
   USHORT        length;
   VM7HDL       free;
   };
typedef struct HandleList VM7HDLLIST;
typedef struct HandleList * VM7HDLLISTPV;

struct VLArray
   {
   SHORT npages;                 /*  number of pages in the array  */
   ULONG nitems;                /*  size of array in bytes  */
   ULONG max;                   /*  highest value used  */
   ULONG current;               /*  current index - makes it list-like  */
   SHORT size;                   /*  the size of each item  */
   SHORT recl;                   /*  the effective record length  */
   VM7HDL hplist;               /*  handle to the page list  */
   USHORT *plist;                /*  pointer to page list - set dynamically 
                                |   when locked  */
   UBYTE bank_id;
   };
typedef struct VLArray VLARRAY;
typedef struct VLArray * VLARRAYPV;

struct memory_vm7
   {
   USHORT block_size;            /* size blocks to be allocated */
   VADDRESS next;               /* ptr to next available byte */
   LONG size;                   /* total size of vm7 */
   LONG avail;                  /* bytes remaining */
   CHAR *swap_file_name;        /* name of swap file for this bank  */
   CHAR *full_name;
   int  swap_file;              /* swap file for paging  */
   SHORT *file_page_list;        /* list of file pages  */
   SHORT n_file_pages;           /* number of initial pages in file  */
   SHORT bump_file_pages;        /* how many to bump it by  */
   CHAR **page_buffer;      /* the page buffer  */
   USHORT n_mem_pages;           /* number of pages in memory  */
   SHORT mem_pages_used;         /* number of pages used in memory  */
   USHORT *page_weight;          /* used in FreePage  */
   USHORT *mem_index;            /* used in FaultPage */
   USHORT *page_locks;           /* used in FreePage  */
   BOOL extended_flag;          /* file extended flag  */
   UBYTE access_flag;           /* READ_WRITE or READ_ONLY  */
   BOOL compress_flag;          /* can/can't compress this bank  */
   SHORT top_write;              /* last page written to disk  */
   VOIDPV ext_rec;              /* extended memory record */
   };
typedef struct memory_vm7 * VM7;

#ifdef DEBUG
static BOOL display_bank[16] = {FALSE, FALSE, FALSE, FALSE, 
                                FALSE, FALSE, FALSE, FALSE,
                                FALSE, FALSE, FALSE, FALSE,
                                FALSE, FALSE, FALSE, FALSE};
VOID xmit_bank(SHORT, BOOL);

#define DRAW_ONE  FALSE
#define DRAW_ALL  TRUE
#endif

VOID DiskWritePage(VM7, SHORT, SHORT); 
VOID DiskReadPage(VM7, SHORT, SHORT);

#ifdef VM7_DEF
#define EXTERN 
VM7 vm7[MAXBANK + 1];
SBYTE cp = DEFAULT_BANK;
CHAR *swap_path = "";           /* swap file path  */
SHORT last_page[MAXBANK + 1];
VOID (*write_page)(VM7, SHORT, SHORT) = DiskWritePage;
VOID (*read_page)(VM7, SHORT, SHORT)  = DiskReadPage;
VOID (*ext_flush_pages)(VM7) = NULL;
VOID (*ext_clean_up)(VM7) = NULL;
SHORT hdl_offset_shift = 1;

#else
#define EXTERN extern
extern VM7 vm7[];
extern SBYTE cp;
extern CHAR *swap_path;         /* swap file path  */
extern SHORT last_page[];
extern VOID (*write_page)(VM7, SHORT, SHORT);
extern VOID (*read_page)(VM7, SHORT, SHORT);
extern VOID (*ext_flush_pages)(VM7);
extern VOID (*ext_clean_up)(VM7);
extern SHORT hdl_offset_shift;

#endif

EXTERN BOOL show_locks;
EXTERN VM7HDL lock_count;
EXTERN VM7HDL stop_lock_handle;
EXTERN VM7HDL stop_alloc_chunk;
EXTERN BOOL ext_init;

#ifdef DEBUG
EXTERN VM7HDLLISTPV db_hlist;
EXTERN VLARRAYPV    db_vlray;
EXTERN VADDRESS  *  db_vadrs;
                        /*  globals for debug  */
#endif
EXTERN VM7HDL handle_count, top_handle;
EXTERN VM7HDL handle_first_free;
EXTERN int zsave_file;

BOOL AllocVm7(VOID);
VM7HDL AddVm7Literal(CHAR *);
void FreeVm7(VOID);
CHAR *GetVm7Base(VOID);
VM7HDL GetVm7Chunk(SHORT);
VM7HDL GetVlarray(ULONG, SHORT);
VLARRAYPV AllocVlarray(ULONG, SHORT);
VOID FreeVlarray(VLARRAYPV);
VOIDPV IndexVm7(VLARRAYPV, ULONG, BOOL);
VOID IndexVm7Unlock(VLARRAYPV, ULONG);
SHORT AddVm7Block(VOID);
FILE *OpenSwapFile(CHAR *);
VOIDPV LockVm7Addrs(VM7HDL, BOOL);
BOOL UnlockVm7Addrs(VM7HDL);
BOOL ReleaseHandle(VM7HDL);
BOOL InitVm7(CHAR *);
VOID DetachSystem(VOID);
BOOL AttachSystem(CHAR *);
VOID CloseVm7(VOID);
VOID DetachVm7(VOID);
BOOL AttachVm7(CHAR *, UBYTE);
VADDRESS GetHandleAddress(VM7HDL);
UBYTE GetHandleID(VM7HDL);
VM7HDL MakeHandle(USHORT, USHORT, USHORT);
VM7HDL ResizeVm7Chunk(VM7HDL, USHORT, USHORT);
VOID SortVlarray(VLARRAYPV, int (*)(), SHORT, SHORT);
VM7HDL BSortVlarray(VLARRAYPV, int(*)(), SHORT);
BOOL VlarrayGet(VLARRAYPV, VOIDPV );
BOOL VlarrayPut(VLARRAYPV, VOIDPV );
VOID HandleDump(VOID);
VOID ResetVm7(VOID);
VOID xmit_string(CHAR *);
BOOL CheckHandle(VM7HDL);
VOID ChangeBankAccess(BOOL);
VOID SaveVm7State(BOOL);
VOID ClearHandles(VOID);
VOID ReduceLast(VOID);
VOID UnlockVm7All(VOID);
BOOL XMS_Init(VOID);
VOID VM7_Error(VOID);
VM7HDL GetVm7First(BOOL);

EXTERN VOID (*VM7_uclean_up)();

#define VM7_Set_SwapPath(a) (swap_path = (a))
#define VM7_Get_SwapPath    (CHAR *)swap_path

#define VM7_Begin_Bank_Def   vm7[cp] = (VM7)Qalloc(sizeof(struct memory_vm7))
#define VM7_Init_System(a)   InitVm7(a)
#define VM7_Close_System     CloseVm7()
#define VM7_Attach_System(a) (BOOL)AttachSystem(a)
#define VM7_Detach_System    DetachSystem()
#define VM7_Reset_Bank       ResetVm7()
#define VM7_Set_Current(a)   (cp) = (UBYTE)a
#define VM7_Get_Current      cp
#define VM7_Init_Bank        AllocVm7()
#define VM7_Free_Bank        FreeVm7()
#define VM7_Flush_Bank       FlushVm7()
#define VM7_Attach_Bank(a,b) (BOOL)AttachVm7((a), (b))
#define VM7_Set_Compress     (vm7[cp]->compress_flag = TRUE)
#define VM7_Set_Debug	     VM7_Set_Compress
#define VM7_Detach_Bank      DetachVm7()
#define VM7_Get_First_Handle GetVm7First(FALSE)
#define VM7_Get_First_VLArray GetVm7First(TRUE)
#define VM7_Set_Page_Size(a) (vm7[cp]->block_size = (a))
#define VM7_Get_Page_Size    (USHORT)(vm7[cp]->block_size)
#define VM7_Set_NPages(a)    (vm7[cp]->n_mem_pages = (a)) 
#define VM7_Get_NPages       ((USHORT)(vm7[cp]->n_mem_pages)
#define VM7_Set_SwapFile(a)  (vm7[cp]->swap_file_name = (a))
#define VM7_Get_SwapFile     (CHAR *)(vm7[cp]->swap_file_name)
#define VM7_Set_NFilePages(a) (vm7[cp]->n_file_pages = (a))
#define VM7_Get_NFilePages   ((USHORT)(vm7[cp]->n_file_pages)
#define VM7_Add_Literal(a)   AddVm7Literal(a)
#define VM7_Lock(a)          LockVm7Addrs((a), TRUE)
#define VM7_Ref(a)           LockVm7Addrs((a), FALSE)
#define VM7_Unlock(a)        (BOOL)UnlockVm7Addrs(a)
#define VM7_Unlock_All       UnlockVm7All()
#define VM7_Make_Handle(a,b,c) (VM7HDL)MakeHandle((USHORT)(a), (USHORT)(b), (USHORT)(c))
#define VM7_Get_Handle_Address(a) (VADDRESS)GetHandleAddress(a)
#define VM7_Get_Handle_ID(a) (UBYTE)GetHandleID(a)
#define VM7_Check_Handle(a)  (BOOL)(CheckHandle(a))
#define VM7_Get_Chunk(a)     (VM7HDL)GetVm7Chunk(a)
#define VM7_Resize_Chunk(a,b,c) (VM7HDL)ResizeVm7Chunk((a), (USHORT)(b), (USHORT)(c))
#define VM7_Free(a)          (BOOL)ReleaseHandle((a))
#define VM7_Get_Array(a,b)   (VM7HDL)(GetVlarray((a), (USHORT)(b)))
#define VM7_Alloc_Array(a,b) (VLARRAYPV)(AllocVlarray((a), (USHORT)(b)))
#define VM7_Free_Array(a)    FreeVlarray(a)
#define VM7_Index_Ref(a,b)   (IndexVm7((a), (USHORT)(b), FALSE))
#define VM7_Index_Lock(a,b)  (IndexVm7((a), (USHORT)(b), TRUE))
#define VM7_Index_Unlock(a,b) (IndexVm7Unlock((a), (USHORT)(b)))
#define VM7_BSort_Array(a,b,c) (VM7HDL)(BSortVlarray((a), (int (*)())(b), (SHORT)(c)))
#define VM7_Array_Top(a)      (a)->current = NULL_ITEM
#define VM7_Array_Next(a)     ((a)->max == NULL_ITEM || (++(a)->current > (a)->max)? FALSE : TRUE)
#define VM7_Array_Last(a)     ((a)->max)
#define VM7_Array_Extend(a)   ((a)->current = (a)->max + 1)
#define VM7_Array_Get(a,b)    VlarrayGet((a),(b))
#define VM7_Array_Put(a,b)    VlarrayPut((a),(b))
#define VM7_Sort_Array(a,b)   (VOID)(SortVlarray((a), (int (*)())(b), -1L, 0L))
#define VM7_LSort_Array(a,b,c,d)  (VOID)(SortVlarray((a), (int (*)())(b), c, d))
#define VM7_DumpHandles       HandleDump()
#define VM7_Set_Access(a)     (ChangeBankAccess(a))

#undef EXTERN
#endif
